from __future__ import absolute_import
from .AIList_core import AIList
from .IntervalArray_core import IntervalArray
from .Interval_core import Interval
from .LabeledIntervalArray_core import LabeledIntervalArray, LabeledInterval, get_include

# This is extracted automatically by the top-level setup.py.
__version__ = '2.1.0'
__author__ = "Kyle S. Smith"


__doc__ = """\
API
======

Basic class
-----------

.. autosummary::
   :toctree: .
   
   Interval
   LabeledInterval
   AIList
   IntervalArray
   LabeledIntervalArray
    
"""